"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from .image import Image


class Jpeg(Image):
    def __init__(self):
        super().__init__("jpeg")



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import sys
from typing import Tuple

import numpy as np

from hub.collections._store_version import CURRENT_STORE_VERSION
from hub.collections._chunk_utils import _tensor_chunksize, _logify_chunksize
from hub.exceptions import ModuleNotInstalledException


def _dask_shape_backward(shape: Tuple[int]):
    if len(shape) == 0:
        return shape
    else:
        return (-1,) + (shape[1:]) if np.isnan(shape[0]) else shape


class Tensor:
    def __init__(self, meta: dict, daskarray, delayed_objs: tuple = None):
        if "dask" not in sys.modules:
            raise ModuleNotInstalledException("dask")
        else:
            import dask
            import dask.array

            global dask

        if not meta.get("preprocessed"):
            meta = Tensor._preprocess_meta(meta, daskarray)
        self._meta = meta
        self._array = daskarray
        self._delayed_objs = delayed_objs
        self._shape = _dask_shape_backward(daskarray.shape)
        self._dtype = meta["dtype"]
        self._dtag = meta.get("dtag")
        self._dcompress = meta.get("dcompress")
        self._dcompress_algo = meta.get("dcompress_algo")
        self._dcompress_lvl = meta.get("dcompress_lvl")
        self._chunksize = meta.get("chunksize")

    @staticmethod
    def _preprocess_meta(meta, daskarray):
        meta = dict(meta)
        meta["preprocessed"] = True
        meta["dtype"] = meta.get("dtype") or daskarray.dtype
        if meta.get("chunksize") is None:
            if str(meta["dtype"]) == "object":
                meta["chunksize"] = 1
            else:
                meta["chunksize"] = _tensor_chunksize(daskarray)
        else:
            meta["chunksize"] = _logify_chunksize(meta["chunksize"])
        meta["shape"] = daskarray.shape
        meta["STORE_VERSION"] = CURRENT_STORE_VERSION
        if "dcompress" in meta:
            dcompress_comp = str(meta["dcompress"]).split(sep=":")
            assert len(dcompress_comp) in [
                1,
                2,
            ], "Invalid dcompress format, should be {algo:compress_lvl} or {algo}"
            meta["dcompress_algo"] = dcompress_comp[0]
            meta["dcompress_lvl"] = (
                dcompress_comp[1] if len(dcompress_comp) == 2 else None
            )
        else:
            meta["dcompress"] = None
            meta["dcompress_algo"] = None
            meta["dcompress_lvl"] = None
        return meta

    @property
    def meta(self):
        """
        Returns
        -------
        dict
            metadata dict of tensor
        """
        return dict(self._meta)

    @property
    def shape(self):
        """
        Returns
        -------
        tuple
            Tensor shape
        """
        return tuple(self._shape)

    @property
    def ndim(self):
        """
        Returns
        -------
        int
            Number of dimensions (len(shape))
        """
        return len(self._shape)

    @property
    def count(self):
        """
        Returns
        -------
        int
            Number of elements on axis 0 if that number is known, -1 otherwise
        """
        return self._shape[0]

    def __len__(self) -> int:
        """
        Returns
        -------
        int
            Number of elements on axis 0 if that number is known, raises Exception otherwise
        """
        if self._shape[0] == -1:
            raise Exception(
                "__len__ can only return >=0 numbers, use .count property if the len of tensor is unknown (in which case -1 is returned)"
            )
        return self._shape[0]

    @property
    def dtype(self):
        """
        Returns
        -------
        str
            Tensor data type, equivalent of numpy dtype
        """
        return self._dtype

    @property
    def dtag(self):
        """
        Returns
        -------
        str
            Information about data stored in tensor (iamge, mask, label, ...)
        """
        return self._dtag

    @property
    def dcompress(self):
        return self._dcompress

    @property
    def chunksize(self):
        return self._chunksize

    def __getitem__(self, slices) -> "Tensor":
        """Slices tensor
        Parameters
        ----------
        slices
            tuple of slices or ints

        Returns
        -------
        Tensor
            sliced tensor
        """
        if self.count == -1:
            raise Exception("Cannot slice an array with unknown len")
        arr = self._array[slices]
        if isinstance(arr, dask.delayed.__class__):
            assert False, "This branch should have never been reached, check me please"
            return arr
        else:
            return Tensor(self._meta, arr)

    def __iter__(self):
        # TODO this function should return tensors, not dask arrays
        for i in range(len(self)):
            yield self._array[i]

    def compute(self):
        """Does lazy computation and converts data to numpy array
        Returns
        -------
        np.ndarray
            numpy array of tensor's data
        """
        return self._array.compute()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import Tuple

from hub.schema.features import Tensor


class Audio(Tensor):
    def __init__(
        self,
        shape: Tuple[int, ...] = (None,),
        dtype="int64",
        file_format=None,
        sample_rate: int = None,
        max_shape: Tuple[int, ...] = None,
        chunks=None,
        compressor="lz4",
    ):
        """Constructs the connector.

        Parameters
        ----------
        file_format: `str`
            the audio file format. Can be any format ffmpeg
            understands. If `None`, will attempt to infer from the file extension.
        shape: `tuple`
             shape of the data.
        dtype: str
            The dtype of the data.
        sample_rate: `int`
            additional metadata exposed to the user through
            `info.schema['audio'].sample_rate`. This value isn't used neither in
            encoding nor decoding.


        Raises
        ----------
        ValueError: If the shape is invalid
        """
        self.file_format = file_format
        if len(shape) != 1:
            raise ValueError(
                f"Audio schema currently only supports 1-D values, got {shape}"
            )
        # self._shape = shape
        self.sample_rate = sample_rate
        super().__init__(
            shape=shape,
            dtype=dtype,
            max_shape=max_shape,
            chunks=chunks,
            compressor=compressor,
        )

    def __str__(self):
        out = super().__str__()
        out = "Audio" + out[6:-1]
        out = (
            out + ", file_format=" + self.file_format
            if self.file_format is not None
            else out
        )
        out = (
            out + ", sample_rate=" + str(self.sample_rate)
            if self.sample_rate is not None
            else out
        )
        out += ")"
        return out

    def __repr__(self):
        return self.__str__()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import os
from pathlib import Path
from hub import config
from hub.log import logger


class TokenManager:
    """ manages access tokens """

    @classmethod
    def is_authenticated(cls):
        return os.path.getsize(config.TOKEN_FILE_PATH) > 10

    @classmethod
    def set_token(cls, token):
        logger.debug(f"Putting the key {token} into {config.TOKEN_FILE_PATH}.")
        path = Path(config.TOKEN_FILE_PATH)
        os.makedirs(path.parent, exist_ok=True)
        with open(config.TOKEN_FILE_PATH, "w") as f:
            f.write(token)

    @classmethod
    def get_token(cls):
        logger.debug("Getting token...")
        if not os.path.exists(config.TOKEN_FILE_PATH):
            return None
        with open(config.TOKEN_FILE_PATH) as f:
            token = f.read()
        logger.debug(f"Got the key {token} from {config.TOKEN_FILE_PATH}.")
        return token

    @classmethod
    def get_auth_header(cls):
        logger.debug("Constructing auth header...")
        token = cls.get_token()
        if token:
            return f"Bearer {token}"
        return None

    @classmethod
    def purge_token(cls):
        logger.debug("Purging token...")
        if os.path.isfile(config.TOKEN_FILE_PATH):
            os.remove(config.TOKEN_FILE_PATH)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
import textwrap

import click
from humbug.report import Report

from hub.log import logger
from hub import config
from hub.client.auth import AuthClient
from hub.client.token_manager import TokenManager
from hub.client.hub_control import HubControlClient
from hub.report import configure_reporting, get_reporting_config, hub_reporter


@click.command()
# @click.option('--token', is_flag=True, default=False,
#        help='Enter authentication tocken from {}'.format(config.GET_TOKEN_REST_SUFFIX))
@click.option("--username", "-u", default=None, help="Your Activeloop AI username")
@click.option("--password", "-p", default=None, help="Your Activeloop AI password")
def login(username, password):
    """ Log in to Activeloop AI"""
    login_fn(username, password)


@click.command()
def logout():
    """ Log out of Activeloop AI"""
    TokenManager.purge_token()


@click.command()
@click.option("--username", "-u", default=None, help="Your Activeloop AI username")
@click.option("--email", "-e", default=None, help="Your email")
@click.option("--password", "-p", default=None, help="Your Activeloop AI password")
def register(username, email, password):
    """ Register at Activeloop AI"""
    if not username:
        logger.debug("Prompting for username.")
        username = click.prompt("Username", type=str)
    username = username.strip()
    if not email:
        logger.debug("Prompting for email.")
        email = click.prompt("Email", type=str)
    email = email.strip()
    if not password:
        logger.debug("Prompting for password.")
        password = click.prompt("Password", type=str, hide_input=True)
    password = password.strip()
    AuthClient().register(username, email, password)
    token = AuthClient().get_access_token(username, password)
    TokenManager.set_token(token)
    consent_message = textwrap.dedent(
        """
        Privacy policy:
        We collect basic system information and crash reports so that we can keep
        improving your experience using Hub to work with your data.

        You can find out more by reading our privacy policy:
            https://www.activeloop.ai/privacy/

        If you would like to opt out of reporting crashes and system information,
        run the following command:
            $ hub reporting --off
        """
    )
    logger.info(consent_message)
    configure_reporting(True, username=username)


@click.command()
@click.option("--on/--off", help="Turn crash report on/off")
def reporting(on):
    """
    Enable or disable sending crash reports to Activeloop AI.
    """
    report = Report(
        title="Consent change",
        tags=hub_reporter.system_tags(),
        content="Consent? `{}`".format(on),
    )
    hub_reporter.publish(report)
    configure_reporting(on)


def login_fn(username, password):
    """ Log in to Activeloop AI"""
    token = ""
    if token:
        logger.info("Token login.")
        logger.degug("Getting the token...")
        token = click.prompt(
            "Please paste the authentication token from {}".format(
                config.GET_TOKEN_REST_SUFFIX, type=str, hide_input=True
            )
        )
        token = token.strip()
        AuthClient.check_token(token)
    else:
        logger.info(
            "Please log in using Activeloop credentials. You can register at https://app.activeloop.ai "
        )
        if not username:
            logger.debug("Prompting for username.")
            username = click.prompt("Username", type=str)
        username = username.strip()
        if not password:
            logger.debug("Prompting for password.")
            password = click.prompt("Password", type=str, hide_input=True)
        password = password.strip()
        token = AuthClient().get_access_token(username, password)
    TokenManager.set_token(token)
    HubControlClient().get_credentials()
    logger.info("Login Successful.")
    reporting_config = get_reporting_config()
    if reporting_config.get("username") != username:
        configure_reporting(True, username=username)



from hub.api.dataset import Dataset, DatasetView, TensorView
import numpy as np


# a list of Datasets or DatasetViews or Tensorviews that supports compute operation
class ComputeList:
    # Doesn't support further get item operations currently
    def __init__(self, ls):
        self.ls = ls

    def compute(self):
        results = [
            item.compute()
            if isinstance(item, (Dataset, DatasetView, TensorView))
            else item
            for item in self.ls
        ]
        return np.concatenate(results)

    def numpy(self):
        return self.compute()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.store.lru_cache import LRUCache

import zarr


def test_lru_cache():
    data = bytes("Hello World", "utf-8")
    cache = LRUCache(zarr.MemoryStore(), zarr.MemoryStore(), 30)
    cache["Aello"] = data
    cache["Beta"] = data
    assert "Aello" in cache._cached_items
    assert "Beta" in cache._cached_items
    assert "Aello" in cache.cache_storage
    assert "Beta" in cache.cache_storage
    cache["Gamma"] = data
    cache["Gamma"] = data
    assert "Aello" not in cache._cached_items
    assert "Aello" not in cache.cache_storage
    assert "Gamma" in cache._cached_items
    assert "Gamma" in cache.cache_storage

    assert list(sorted(cache)) == ["Aello", "Beta", "Gamma"]
    assert list(sorted(cache.cache_storage)) == ["Beta", "Gamma"]
    assert list(sorted(cache.actual_storage)) == ["Aello"]
    del cache["Gamma"]
    assert list(sorted(cache)) == ["Aello", "Beta"]
    assert list(sorted(cache.cache_storage)) == ["Beta"]
    cache["Aello"]
    cache["Beta"]
    try:
        del cache["KeyError"]
    except KeyError:
        pass
    assert list(sorted(cache.actual_storage)) == ["Aello"]
    cache.flush()
    assert list(sorted(cache.actual_storage)) == ["Aello", "Beta"]
    cache.commit()


if __name__ == "__main__":
    test_lru_cache()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.utils import _tuple_product
from hub.api.tensorview import TensorView
import collections.abc as abc
from hub.api.dataset_utils import (
    create_numpy_dict,
    get_value,
    slice_split,
    str_to_int,
)
from hub.exceptions import NoneValueException
from hub.api.objectview import ObjectView
from hub.schema import Sequence
import numpy as np


class DatasetView:
    def __init__(
        self,
        dataset=None,
        lazy: bool = True,
        indexes=None,  # list or integer
    ):
        """Creates a DatasetView object for a subset of the Dataset.

        Parameters
        ----------
        dataset: hub.api.dataset.Dataset object
            The dataset whose DatasetView is being created
        lazy: bool, optional
            Setting this to False will stop lazy computation and will allow items to be accessed without .compute()
        indexes: optional
            It can be either a list or an integer depending upon the slicing. Represents the indexes that the datasetview is representing.
        """
        if dataset is None:
            raise NoneValueException("dataset")
        if indexes is None:
            raise NoneValueException("indexes")

        self.dataset = dataset
        self.lazy = lazy
        self.indexes = indexes
        self.is_contiguous = False
        if isinstance(self.indexes, list) and self.indexes:
            self.is_contiguous = self.indexes[-1] - self.indexes[0] + 1 == len(
                self.indexes
            )

    def __getitem__(self, slice_):
        """| Gets a slice or slices from DatasetView
        | Usage:

        >>> ds_view = ds[5:15]
        >>> return ds_view["image", 7, 0:1920, 0:1080, 0:3].compute() # returns numpy array of 12th image
        """
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        slice_list = [0] + slice_list if isinstance(self.indexes, int) else slice_list
        if not subpath:
            if len(slice_list) > 1:
                raise ValueError("Can't slice dataset with multiple slices without key")
            indexes = self.indexes[slice_list[0]]
            return DatasetView(dataset=self.dataset, lazy=self.lazy, indexes=indexes)
        elif not slice_list:
            slice_ = (
                [slice(self.indexes[0], self.indexes[-1] + 1)]
                if self.is_contiguous
                else [self.indexes]
            )
            if subpath in self.keys:
                tensorview = TensorView(
                    dataset=self.dataset,
                    subpath=subpath,
                    slice_=slice_,
                    lazy=self.lazy,
                )
                return tensorview if self.lazy else tensorview.compute()
            for key in self.keys:
                if subpath.startswith(key):
                    objectview = ObjectView(
                        dataset=self.dataset,
                        subpath=subpath,
                        slice_=slice_,
                        lazy=self.lazy,
                    )
                    return objectview if self.lazy else objectview.compute()
            return self._get_dictionary(subpath, slice_)
        else:
            if isinstance(self.indexes, list):
                indexes = self.indexes[slice_list[0]]
                if self.is_contiguous and isinstance(indexes, list) and indexes:
                    indexes = slice(indexes[0], indexes[-1] + 1)
            else:
                indexes = self.indexes
            slice_list[0] = indexes
            schema_obj = self.dataset.schema.dict_[subpath.split("/")[1]]

            if subpath in self.keys and (
                not isinstance(schema_obj, Sequence) or len(slice_list) <= 1
            ):
                tensorview = TensorView(
                    dataset=self.dataset,
                    subpath=subpath,
                    slice_=slice_list,
                    lazy=self.lazy,
                )
                return tensorview if self.lazy else tensorview.compute()
            for key in self.keys:
                if subpath.startswith(key):
                    objectview = ObjectView(
                        dataset=self.dataset,
                        subpath=subpath,
                        slice_=slice_list,
                        lazy=self.lazy,
                    )
                    return objectview if self.lazy else objectview.compute()
            if len(slice_list) > 1:
                raise ValueError("You can't slice a dictionary of Tensors")
            return self._get_dictionary(subpath, slice_list[0])

    def __setitem__(self, slice_, value):
        """| Sets a slice or slices with a value
        | Usage:

        >>> ds_view = ds[5:15]
        >>> ds_view["image", 3, 0:1920, 0:1080, 0:3] = np.zeros((1920, 1080, 3), "uint8") # sets the 8th image
        """
        assign_value = get_value(value)
        assign_value = str_to_int(
            assign_value, self.dataset.tokenizer
        )  # handling strings and bytes

        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        slice_list = [0] + slice_list if isinstance(self.indexes, int) else slice_list

        if not subpath:
            raise ValueError("Can't assign to dataset sliced without key")
        elif subpath not in self.keys:
            raise KeyError(f"Key {subpath} not found in dataset")

        if not slice_list:
            slice_ = (
                slice(self.indexes[0], self.indexes[-1] + 1)
                if self.is_contiguous
                else self.indexes
            )
            if not isinstance(slice_, list):
                self.dataset._tensors[subpath][slice_] = assign_value
            else:
                for i, index in enumerate(slice_):
                    self.dataset._tensors[subpath][index] = assign_value[i]
        else:
            if isinstance(self.indexes, list):
                indexes = self.indexes[slice_list[0]]
                if self.is_contiguous and isinstance(indexes, list) and indexes:
                    slice_list[0] = slice(indexes[0], indexes[-1] + 1)
                else:
                    slice_list[0] = indexes
            else:
                slice_list[0] = self.indexes

            if not isinstance(slice_list[0], list):
                self.dataset._tensors[subpath][slice_list] = assign_value
            else:
                for i, index in enumerate(slice_list[0]):
                    current_slice = [index] + slice_list[1:]
                    self.dataset._tensors[subpath][current_slice] = assign_value[i]

    def filter(self, fn):
        """| Applies a function on each element one by one as a filter to get a new DatasetView

        Parameters
        ----------
        fn: function
            Should take in a single sample of the dataset and return True or False
            This function is applied to all the items of the datasetview and retains those items that return True
        """
        indexes = []
        if isinstance(self.indexes, int):
            dsv = self.dataset[self.indexes]
            if fn(dsv):
                return DatasetView(
                    dataset=self.dataset, lazy=self.lazy, indexes=self.indexes
                )
        else:
            indexes = [index for index in self.indexes if fn(self.dataset[index])]
        return DatasetView(dataset=self.dataset, lazy=self.lazy, indexes=indexes)

    @property
    def keys(self):
        """
        Get Keys of the dataset
        """
        return self.dataset._tensors.keys()

    @property
    def schema(self):
        return self.dataset.schema

    def _get_dictionary(self, subpath, slice_):
        """Gets dictionary from dataset given incomplete subpath"""
        tensor_dict = {}
        subpath = subpath if subpath.endswith("/") else subpath + "/"
        for key in self.keys:
            if key.startswith(subpath):
                suffix_key = key[len(subpath) :]
                split_key = suffix_key.split("/")
                cur = tensor_dict
                for sub_key in split_key[:-1]:
                    if sub_key not in cur.keys():
                        cur[sub_key] = {}
                    cur = cur[sub_key]
                tensorview = TensorView(
                    dataset=self.dataset,
                    subpath=key,
                    slice_=slice_,
                    lazy=self.lazy,
                )
                cur[split_key[-1]] = tensorview if self.lazy else tensorview.compute()
        if not tensor_dict:
            raise KeyError(f"Key {subpath} was not found in dataset")
        return tensor_dict

    def __iter__(self):
        """ Returns Iterable over samples """
        if isinstance(self.indexes, int):
            yield self
            return

        for i in range(len(self.indexes)):
            yield self[i]

    def __len__(self):
        return len(self.indexes) if isinstance(self.indexes, list) else 1

    def __str__(self):
        return "DatasetView(" + str(self.dataset) + ")"

    def __repr__(self):
        return self.__str__()

    def to_tensorflow(self, include_shapes):
        """|Converts the dataset into a tensorflow compatible format

        Parameters
        ----------
        include_shapes: boolean, optional
            False by deefault. Setting it to True passes the shapes to tf.data.Dataset.from_generator.
            Setting to True could lead to issues with dictionaries inside Tensors.
        """

        return self.dataset.to_tensorflow(
            indexes=self.indexes, include_shapes=include_shapes
        )

    def to_pytorch(
        self,
        transform=None,
        inplace=True,
        output_type=dict,
    ):
        """| Converts the dataset into a pytorch compatible format.

        Parameters
        ----------
        transform: function that transforms data in a dict format
        inplace: bool, optional
            Defines if data should be converted to torch.Tensor before or after Transforms applied (depends on what data
            type you need for Transforms). Default is True.
        output_type: one of list, tuple, dict, optional
            Defines the output type. Default is dict - same as in original Hub Dataset.
        """
        return self.dataset.to_pytorch(
            transform=transform,
            indexes=self.indexes,
            inplace=inplace,
            output_type=output_type,
        )

    def resize_shape(self, size: int) -> None:
        """Resize dataset shape, not DatasetView"""
        self.dataset.resize_shape(size)

    def commit(self) -> None:
        """Commit dataset"""
        self.dataset.commit()

    def numpy(self, label_name=False):
        """Gets the value from different tensorview objects in the datasetview schema

        Parameters
        ----------
        label_name: bool, optional
            If the TensorView object is of the ClassLabel type, setting this to True would retrieve the label names
            instead of the label encoded integers, otherwise this parameter is ignored.
        """
        if isinstance(self.indexes, int):
            return create_numpy_dict(self.dataset, self.indexes, label_name=label_name)
        else:
            return np.array(
                [
                    create_numpy_dict(self.dataset, index, label_name=label_name)
                    for index in self.indexes
                ]
            )

    def disable_lazy(self):
        self.lazy = False

    def enable_lazy(self):
        self.lazy = True

    def compute(self, label_name=False):
        """Gets the value from different tensorview objects in the datasetview schema

        Parameters
        ----------
        label_name: bool, optional
            If the TensorView object is of the ClassLabel type, setting this to True would retrieve the label names
            instead of the label encoded integers, otherwise this parameter is ignored.
        """
        return self.numpy(label_name=label_name)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import Tuple
import numpy as np

from hub.schema.features import Tensor
from hub.schema.class_label import ClassLabel


class Segmentation(Tensor):
    """`HubSchema` for segmentation"""

    def __init__(
        self,
        shape: Tuple[int, ...] = None,
        dtype: str = None,
        num_classes: int = None,
        names: Tuple[str] = None,
        names_file: str = None,
        max_shape: Tuple[int, ...] = None,
        chunks=None,
        compressor="lz4",
    ):
        """Constructs a Segmentation HubSchema.
        Also constructs ClassLabel HubSchema for Segmentation classes.

        Parameters
        ----------
        shape: tuple of ints or None
            Shape in format (height, width, 1)
        dtype: str
            dtype of segmentation array: `uint16` or `uint8`
        num_classes: int
            Number of classes. All labels must be < num_classes.
        names: `list<str>`
            string names for the integer classes. The order in which the names are provided is kept.
        names_file: str
            Path to a file with names for the integer classes, one per line.
        max_shape : tuple[int]
            Maximum shape of tensor shape if tensor is dynamic
        chunks : tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks
        """
        super().__init__(shape, dtype, max_shape=max_shape, chunks=chunks)
        self.class_labels = ClassLabel(
            num_classes=num_classes,
            names=names,
            names_file=names_file,
            chunks=chunks,
            compressor="lz4",
        )

    def get_segmentation_classes(self):
        """Get classes of the segmentation mask"""
        return self.class_labels.names

    def __str__(self):
        out = super().__str__()
        out = "Segmentation" + out[6:-1]
        out = (
            out + ", names=" + str(self.class_labels._names)
            if self.class_labels._names is not None
            else out
        )
        out = (
            out + ", num_classes=" + str(self.class_labels._num_classes)
            if self.class_labels._num_classes is not None
            else out
        )
        out += ")"
        return out

    def __repr__(self):
        return self.__str__()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import Tuple

import numpy as np

from hub.schema.features import Tensor


class Text(Tensor):
    """`HubSchema` for text"""

    def __init__(
        self,
        shape: Tuple[int, ...] = (None,),
        dtype="int64",
        max_shape: Tuple[int, ...] = None,
        chunks=None,
        compressor="lz4",
    ):
        """| Construct the connector.
        Returns integer representation of given string.

        Parameters
        ----------
        shape: tuple of ints or None
            The shape of the text
        dtype: str
            the dtype for storage.
        max_shape : Tuple[int]
            Maximum number of words in the text
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks
        """
        self._set_dtype(dtype)
        super().__init__(
            shape,
            dtype,
            max_shape=max_shape,
            chunks=chunks,
            compressor=compressor,
        )

    def _set_dtype(self, dtype):
        """Set the dtype."""
        dtype = str(np.dtype(dtype))
        self.dtype = dtype

    def __str__(self):
        out = super().__str__()
        out = "Text" + out[6:]
        return out

    def __repr__(self):
        return self.__str__()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import os
import json
import uuid

from humbug.consent import HumbugConsent, environment_variable_opt_out
from humbug.report import Reporter

from hub.config import (
    HUMBUG_TOKEN,
    HUMBUG_KB_ID,
    REPORT_CONSENT_FILE_PATH,
)
from hub.version import __version__ as hub_version


def hub_consent_from_file() -> bool:
    consent = False
    try:
        if not os.path.isfile(REPORT_CONSENT_FILE_PATH):
            return False
        consent_config = {}
        with open(REPORT_CONSENT_FILE_PATH, "r") as ifp:
            consent_config = json.load(ifp)
        consent = consent_config.get("consent", False)
    except Exception:
        pass
    return consent


def configure_reporting(consent, client_id=None, username=None):
    """
    Allow or disallow Hub reporting.
    """
    reporting_config = {}
    if os.path.isfile(REPORT_CONSENT_FILE_PATH):
        try:
            with open(REPORT_CONSENT_FILE_PATH, "r") as ifp:
                reporting_config = json.load(ifp)
        except Exception:
            pass

    if client_id is not None and reporting_config.get("client_id") is None:
        reporting_config["client_id"] = client_id

    if reporting_config.get("client_id") is None:
        reporting_config["client_id"] = str(uuid.uuid4())

    if username is not None:
        reporting_config["username"] = username

    reporting_config["consent"] = consent

    try:
        config_dir = os.path.dirname(REPORT_CONSENT_FILE_PATH)
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)

        with open(REPORT_CONSENT_FILE_PATH, "w") as ofp:
            json.dump(reporting_config, ofp)
    except Exception:
        pass


def get_reporting_config():
    reporting_config = {}
    try:
        if not os.path.exists(REPORT_CONSENT_FILE_PATH):
            client_id = str(uuid.uuid4())
            reporting_config["client_id"] = client_id
            configure_reporting(True, client_id)
        with open(REPORT_CONSENT_FILE_PATH, "r") as ifp:
            reporting_config = json.load(ifp)
    except Exception:
        pass
    return reporting_config


session_id = str(uuid.uuid4())

hub_consent = HumbugConsent(hub_consent_from_file)
hub_reporter = Reporter(
    "activeloopai/Hub",
    hub_consent,
    client_id=get_reporting_config().get("client_id"),
    session_id=session_id,
    bugout_token=HUMBUG_TOKEN,
    bugout_journal_id=HUMBUG_KB_ID,
    timeout_seconds=5,
)


hub_version_tag = "version:{}".format(hub_version)
hub_tags = [hub_version_tag]


class ExceptionWithReporting(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        hub_reporter.error_report(self, tags=hub_tags, publish=True)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import math

import numpy as np


def _logify_chunksize(chunksize):
    return 2 ** int(math.ceil(math.log2(max(chunksize, 1))))


def _tensor_chunksize(t: np.ndarray):
    sz = np.dtype(t.dtype).itemsize * np.prod(t.shape[1:])
    return _logify_chunksize((2 ** 24) / sz)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.schema import Sequence, Tensor, SchemaDict, Primitive
from hub.api.dataset_utils import slice_extract_info, slice_split

import collections.abc as abc


class ObjectView:
    def __init__(
        self,
        dataset,
        subpath=None,
        slice_=None,
        indexes=None,
        nums=[],
        offsets=[],
        squeeze_dims=[],
        inner_schema_obj=None,
        lazy=True,
        check_bounds=True,
    ):
        """Creates an ObjectView object for dataset from a Dataset, DatasetView or TensorView
        object, or creates a different ObjectView from an existing one

        Parameters
        ----------
        These parameters are used to create a new ObjectView.
        dataset: hub.api.dataset.Dataset object
            The dataset whose ObjectView is being created, or its DatasetView
        subpath: str (optional)
            A potentially incomplete path to any element in the Dataset
        slice_list: optional
            The `slice_` of this Tensor that needs to be accessed
        lazy: bool, optional
            Setting this to False will stop lazy computation and will allow items to be accessed without .compute()

        These parameters are also needed to create an ObjectView from an existing one.
        nums: List[int]
            Number of elements in each dimension of the ObjectView to be created
        offsets: List[int]
            Starting element in each dimension of the ObjectView to be created
        squeeze_dims: List[bool]
            Whether each dimension can be squeezed or not
        inner_schema_obj: Child of hub.schema.Tensor or hub.schema.SchemaDict
            The deepest element in the schema upto which the previous ObjectView had been processed
        check_bounds: bool
            Whether to create a new ObjectView object from a Dataset, DatasetView or TensorView
            or create a different ObjectView from an existing one
        """
        self.dataset = dataset
        self.schema = dataset.schema.dict_
        self.subpath = subpath

        self.nums = nums
        self.offsets = offsets
        self.squeeze_dims = squeeze_dims

        self.inner_schema_obj = inner_schema_obj
        self.lazy = lazy

        if check_bounds:
            if self.subpath:
                (
                    self.inner_schema_obj,
                    self.nums,
                    self.offsets,
                    self.squeeze_dims,
                ) = self.process_path(
                    self.subpath,
                    self.inner_schema_obj,
                    self.nums.copy(),
                    self.offsets.copy(),
                    self.squeeze_dims.copy(),
                )
            if slice_ and len(slice_) >= 1:
                self.indexes = slice_[0]
                self.is_contiguous = False
                if isinstance(self.indexes, list) and self.indexes:
                    self.is_contiguous = self.indexes[-1] - self.indexes[0] + 1 == len(
                        self.indexes
                    )
                slice_ = slice_[1:]
                if len(slice_) > len(self.nums):
                    raise IndexError("Too many indices")
                for i, it in enumerate(slice_):
                    num, ofs = slice_extract_info(it, self.nums[i])
                    self.nums[i] = num
                    self.offsets[i] += ofs
                    self.squeeze_dims[i] = num == 1
        else:
            self.indexes = indexes
            self.is_contiguous = False
            if isinstance(self.indexes, list) and self.indexes:
                self.is_contiguous = self.indexes[-1] - self.indexes[0] + 1 == len(
                    self.indexes
                )

    def num_process(self, schema_obj, nums, offsets, squeeze_dims):
        """Determines the maximum number of elements in each discovered dimension"""
        if isinstance(schema_obj, SchemaDict):
            return
        elif isinstance(schema_obj, Sequence):
            nums.append(0)
            offsets.append(0)
            squeeze_dims.append(False)
            if isinstance(schema_obj.dtype, Tensor):
                self.num_process(schema_obj.dtype, nums, offsets, squeeze_dims)
        else:
            for dim in schema_obj.max_shape:
                nums.append(dim)
                offsets.append(0)
                squeeze_dims.append(False)
        if not isinstance(schema_obj.dtype, Primitive) and not isinstance(
            schema_obj, Sequence
        ):
            raise ValueError("Only sequences can be nested")

    def process_path(self, subpath, inner_schema_obj, nums, offsets, squeeze_dims):
        """Checks if a subpath is valid or not. Does not repeat computation done in a previous ObjectView object"""
        paths = subpath.split("/")[1:]
        try:
            # If key is invalid raises KeyError
            # If schema object is not subscriptable raises AttributeError
            if inner_schema_obj:
                if isinstance(inner_schema_obj, Sequence):
                    schema_obj = inner_schema_obj.dtype.dict_[paths[0]]
                elif isinstance(inner_schema_obj, SchemaDict):
                    schema_obj = inner_schema_obj.dict_[paths[0]]
                else:
                    raise KeyError()
            else:
                schema_obj = self.schema[paths[0]]
        except (KeyError, AttributeError):
            raise KeyError(f"{paths[0]} is an invalid key")
        self.num_process(schema_obj, nums, offsets, squeeze_dims)
        for path in paths[1:]:
            try:
                if isinstance(schema_obj, Sequence):
                    schema_obj = schema_obj.dtype.dict_[path]
                elif isinstance(schema_obj, SchemaDict):
                    schema_obj = schema_obj.dict_[path]
                else:
                    raise KeyError()
                self.num_process(schema_obj, nums, offsets, squeeze_dims)
            except (KeyError, AttributeError):
                raise KeyError(f"{path} is an invalid key")
        return schema_obj, nums, offsets, squeeze_dims

    def __getitem__(self, slice_):
        """| Gets a slice from an objectview"""
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)

        nums, offsets, squeeze_dims, inner_schema_obj = (
            self.nums.copy(),
            self.offsets.copy(),
            self.squeeze_dims.copy(),
            self.inner_schema_obj,
        )

        if subpath:
            inner_schema_obj, nums, offsets, squeeze_dims = self.process_path(
                subpath, inner_schema_obj, nums, offsets, squeeze_dims
            )
        subpath = self.subpath + subpath

        new_indexes = self.indexes
        if len(slice_list) >= 1:
            if isinstance(self.indexes, list):
                new_indexes = self.indexes[slice_list[0]]
                if self.is_contiguous and new_indexes:
                    new_indexes = slice(new_indexes[0], new_indexes[-1] + 1)
                slice_list = slice_list[1:]
            elif isinstance(self.indexes, slice):
                ofs = self.indexes.start or 0
                num = self.indexes.stop - ofs if self.indexes.stop else None
                num, ofs_temp = slice_extract_info(slice_list[0], num)
                new_indexes = (
                    ofs + ofs_temp
                    if isinstance(slice_list[0], int)
                    else slice(ofs + ofs_temp, ofs + ofs_temp + num)
                )
                slice_list = slice_list[1:]

        if len(slice_list) >= 1:
            # Expand slice list
            exp_slice_list = []
            for squeeze in squeeze_dims:
                if squeeze:
                    exp_slice_list += [None]
                else:
                    if len(slice_list) > 0:
                        exp_slice_list += [slice_list.pop(0)]
                    else:
                        # slice list smaller than max
                        exp_slice_list += [None]
            if len(slice_list) > 0:
                # slice list longer than max
                raise IndexError("Too many indices")
            for i, it in enumerate(exp_slice_list):
                if it is not None:
                    num, ofs = slice_extract_info(it, nums[i])
                    nums[i] = num
                    offsets[i] += ofs
                    squeeze_dims[i] = isinstance(it, int)

        objectview = ObjectView(
            dataset=self.dataset,
            subpath=subpath,
            slice_=None,
            indexes=new_indexes,
            nums=nums,
            offsets=offsets,
            squeeze_dims=squeeze_dims,
            inner_schema_obj=inner_schema_obj,
            lazy=self.lazy,
            check_bounds=False,
        )
        return objectview if self.lazy else objectview.compute()

    def numpy(self):
        """Gets the value from the objectview"""
        if isinstance(self.indexes, list):
            if len(self.indexes) > 1:
                raise IndexError("Can only go deeper on single datapoint")
            else:
                slice_0 = self.indexes[0]
        elif isinstance(self.indexes, slice):
            if self.indexes.stop - self.indexes.start > 1:
                raise IndexError("Can only go deeper on single datapoint")
            slice_0 = self.indexes.start
        else:
            slice_0 = self.indexes
        # single datapoint
        paths = self.subpath.split("/")[1:]
        schema = self.schema[paths[0]]
        slice_ = [
            ofs if sq else slice(ofs, ofs + num) if num else slice(None, None)
            for ofs, num, sq in zip(self.offsets, self.nums, self.squeeze_dims)
        ]
        if isinstance(schema, Sequence):
            if isinstance(schema.dtype, SchemaDict):
                # if sequence of dict, have to fetch everything
                lazy = self.dataset.lazy
                self.dataset.lazy = False
                value = self.dataset[[paths[0], slice_0]]
                self.dataset.lazy = lazy
                for path in paths[1:]:
                    value = value[path]
                try:
                    return value[tuple(slice_)]
                except TypeError:
                    # raise error
                    return value
                except KeyError:
                    raise KeyError("Invalid slice")
            else:
                # sequence of tensors
                return self.dataset[[paths[0], slice_0]].compute()[tuple(slice_)]

    def compute(self):
        return self.numpy()

    def __str__(self):
        slice_ = [
            ofs if sq else slice(ofs, ofs + num) if num else slice(None, None)
            for ofs, num, sq in zip(self.offsets, self.nums, self.squeeze_dims)
        ]
        return f"ObjectView(subpath='{self.subpath}', indexes={str(self.indexes)}, slice={str(slice_)})"



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from collections import OrderedDict
from collections.abc import MutableMapping


class DummyLock:
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


Lock = DummyLock


class LRUCache(MutableMapping):
    def __init__(
        self,
        cache_storage: MutableMapping,
        actual_storage: MutableMapping,
        max_size,
    ):
        """Creates LRU cache using cache_storage and actual_storage containers
        max_size -> maximum cache size that is allowed
        """
        self._dirty = set()
        self._mutex = Lock()
        self._max_size = max_size
        self._cache_storage = cache_storage
        self._actual_storage = actual_storage
        self._total_cached = 0
        self._cached_items = OrderedDict()
        # assert len(self._cache_storage) == 0, "Initially cache storage should be empty"

    @property
    def cache_storage(self):
        """Storage which is used for caching
        Returns MutableMapping
        """
        return self._cache_storage

    @property
    def actual_storage(self):
        """Storage which is used for actual storing (not caching)
        Returns MutableMapping
        """
        return self._actual_storage

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def _flush_dirty(self):
        for item in self._dirty:
            self._actual_storage[item] = self._cache_storage[item]
        self._dirty.clear()

    def flush(self):
        self._flush_dirty()
        if hasattr(self._cache_storage, "flush"):
            self._cache_storage.flush()
        if hasattr(self._actual_storage, "flush"):
            self._actual_storage.flush()

    def close(self):
        self._flush_dirty()
        if hasattr(self._cache_storage, "close"):
            self._cache_storage.close()
        if hasattr(self._actual_storage, "close"):
            self._actual_storage.close()

    def commit(self):
        self.close()

    def __getitem__(self, key):
        """ Gets item and puts it in the cache if not there """
        with self._mutex:
            if key in self._cached_items:
                self._cached_items.move_to_end(key)
                return self._cache_storage[key]
            else:
                result = self._actual_storage[key]
                self._free_memory(len(result))
                self._append_cache(key, result)
                return result

    def __setitem__(self, key, value):
        """ Sets item and puts it in the cache if not there"""
        with self._mutex:
            if key in self._cached_items:
                self._total_cached -= self._cached_items.pop(key)
            self._free_memory(len(value))
            self._append_cache(key, value)
            if key not in self._dirty:
                self._dirty.add(key)

    def __delitem__(self, key):
        deleted_from_cache = False
        with self._mutex:
            if key in self._cached_items:
                self._total_cached -= self._cached_items.pop(key)
                del self._cache_storage[key]
                self._dirty.discard(key)
                deleted_from_cache = True
            try:
                del self._actual_storage[key]
            except KeyError:
                if not deleted_from_cache:
                    raise

    def __len__(self):
        return len(
            self.actual_storage
        )  # TODO: In future might need to fix this to return proper len

    def __iter__(self):
        cached_keys = set(self._dirty)
        for i in self.actual_storage:
            cached_keys.discard(i)
            yield i
        yield from sorted(cached_keys)

    def _free_memory(self, extra_size):
        while (
            self._total_cached > 0 and extra_size + self._total_cached > self._max_size
        ):
            item, itemsize = self._cached_items.popitem(last=False)
            if item in self._dirty:
                self._actual_storage[item] = self._cache_storage[item]
                self._dirty.discard(item)
            del self._cache_storage[item]
            self._total_cached -= itemsize

    def _append_cache(self, key, value):
        self._total_cached += len(value)
        self._cached_items[key] = len(value)
        self._cache_storage[key] = value



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.schema.features import Tensor


class Sequence(Tensor):
    """`Sequence` correspond to sequence of `features.HubSchema`.
    At generation time, a list for each of the sequence element is given. The output
    of `Dataset` will batch all the elements of the sequence together.
    If the length of the sequence is static and known in advance, it should be
    specified in the constructor using the `length` param.

    | Usage:
    ----------

    >>> sequence = Sequence(Image(), length=NB_FRAME)
    """

    def __init__(
        self,
        shape=(),
        max_shape=None,
        dtype=None,
        chunks=None,
        compressor="lz4",
    ):
        """| Construct a sequence of Tensors.

        Parameters
        ----------
        shape : Tuple[int] | int
            Single integer element tuple representing length of sequence
            If None then dynamic
        dtype : str | HubSchema
            Datatype of each element in sequence
        chunks : Tuple[int] | int
            Number of elements in chunk
            Works only for top level sequence
            You can also include number of samples in a single chunk
        """
        super().__init__(
            shape=shape,
            dtype=dtype,
            chunks=chunks,
            compressor=compressor,
            max_shape=max_shape,
        )

    def __str__(self):
        out = super().__str__()
        out = "Sequence" + out[6:]
        return out

    def __repr__(self):
        return self.__str__()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

# from typing import Tuple

from hub.schema.features import Tensor


class BBox(Tensor):
    """| HubSchema` for a normalized bounding box.

    Output: Tensor of type `float32` and shape `[4,]` which contains the
    normalized coordinates of the bounding box `[xmin, ymin, xmax, ymax]`
    """

    def __init__(self, dtype="float64", chunks=None, compressor="lz4"):
        """Construct the connector.

        Parameters
        ----------
        dtype : str
                dtype of bbox coordinates. Default: 'float32'
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks
        """
        super(BBox, self).__init__(
            shape=(4,), dtype=dtype, chunks=chunks, compressor=compressor
        )

    def __str__(self):
        out = super().__str__()
        out = "BBox" + out[6:]
        return out

    def __repr__(self):
        return self.__str__()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import itertools
from typing import Iterable, Tuple, Dict
import sys

import numpy as np

from .core import Dataset, DatasetGenerator, load, _dask_shape, get_text
from hub.collections.tensor import Tensor


def _generate(generator: DatasetGenerator, input):
    output = generator(input)
    keys = sorted(output.keys())
    return [output[key] for key in keys]


def _meta_preprocess(meta: dict):
    meta = dict(meta)
    assert (
        len({v["shape"][0] for v in meta.values()}) == 1
    ), "first dim should be equal for all tensors"
    for key, value in meta.items():
        assert "dtype" in value
        assert "shape" in value
    return meta


# def _generate_input_preprocess(input):
#     if isinstance(input, Tensor):
#         return input._array
#     else:
#         return input


def generate(generator: DatasetGenerator, input) -> Dataset:
    """Generates dataset based on DatabaseGenerator class instance and iterable input
    For every element in input runs generators __call__ function.
    That function should return dict of numpy arrays containing single or multiple outputs for axis 0 of generating dataset
    """
    if "dask" not in sys.modules:
        raise ModuleNotInstalledException("dask")
    else:
        import dask
        import dask.array

        global dask

    meta = _meta_preprocess(generator.meta())
    keys = sorted(meta.keys())

    tasks = [dask.delayed(_generate, nout=len(meta))(generator, i) for i in input]
    if len(tasks) == 0:
        return Dataset(
            {
                key: Tensor(
                    meta[key],
                    dask.array.from_array(np.empty(shape=(0,), dtype="uint8")),
                )
                for ki, key in enumerate(keys)
            }
        )

    return Dataset(
        {
            key: Tensor(
                meta[key],
                dask.array.concatenate(
                    [
                        dask.array.from_delayed(
                            task[ki],
                            shape=_dask_shape(meta[key]["shape"]),
                            dtype=meta[key]["dtype"],
                        )
                        for task in tasks
                    ]
                ),
                delayed_objs=[task[ki] for task in tasks],
            )
            for ki, key in enumerate(keys)
        }
    )


def from_tensors(
    tensors: dict,
    license: str = None,
    description: str = None,
    citation: str = None,
    howtoload: str = None,
) -> Dataset:
    """Creates a dataset from dict of tensors"""
    return Dataset(
        tensors,
        metainfo={
            "license": license,
            "description": description,
            "citation": citation,
            "howtoload": howtoload,
        },
    )


def _meta_concat(metas: Tuple[Dict[str, object]]):
    _meta = metas[0]
    for meta in metas:
        assert _meta["dtype"] == meta["dtype"]
        assert _meta["shape"][1:] == meta["shape"][1:]
        assert _meta.get("dtag") == meta.get("dtag")
        assert _meta["dcompress"] == meta["dcompress"]

    _meta["shape"] = (sum([meta["shape"][0] for meta in metas]),) + _meta["shape"][1:]
    if _meta["shape"][0] < 0:
        _meta["shape"][0] = -1

    return _meta


def concat(datasets: Iterable[Dataset]) -> Dataset:
    """Concats multiple datasets into one along axis 0
    This is equivalent to concat every tensor with the same key
    """
    if "dask" not in sys.modules:
        raise ModuleNotInstalledException("dask")
    else:
        import dask
        import dask.array

        global dask

    keys = [sorted(dataset._tensors.keys()) for dataset in datasets]
    for key in keys:
        assert key == keys[0]
    keys = keys[0]
    return Dataset(
        {
            key: Tensor(
                _meta_concat([dataset._tensors[key]._meta for dataset in datasets]),
                dask.array.concatenate(
                    [dataset._tensors[key]._array for dataset in datasets]
                ),
                tuple(
                    itertools.chain(
                        *[
                            dataset._tensors[key]._delayed_objs or []
                            for dataset in datasets
                        ]
                    )
                ),
            )
            for key in keys
        }
    )


def merge(datasets: Iterable[Dataset]) -> Dataset:
    """Merges multiple datasets that have distinct keys into one big datasets containing all keys"""
    tensors = {}
    for dataset in datasets:
        for tname, tvalue in dataset._tensors.items():
            assert tname not in tensors
            tensors[tname] = tvalue
    return Dataset(tensors)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import numpy


class Base:
    def encode(self, array: numpy.ndarray) -> bytes:
        raise NotImplementedError()

    def decode(self, bytes: bytes) -> numpy.ndarray:
        raise NotImplementedError()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from io import BytesIO

import zarr
import numcodecs
from numcodecs.abc import Codec
import numpy as np
from PIL import Image


class PngCodec(Codec):
    def __init__(self, solo_channel=True):
        self.codec_id = "png"
        self.solo_channel = solo_channel
        self._msgpack = numcodecs.MsgPack()

    def encode_single_image(self, image: np.ndarray) -> bytes:
        with BytesIO() as buffer:
            Image.fromarray(image).save(buffer, format="png")
            return buffer.getvalue()

    def decode_single_image(self, buf) -> np.ndarray:
        with BytesIO(buf) as buffer:
            buffer.seek(0)
            return np.array(Image.open(buffer, mode="r"))

    def encode(self, buf: np.ndarray):
        append_one = False
        if self.solo_channel and buf.shape[-1] == 1:
            buf = np.reshape(buf, buf.shape[:-1])
            append_one = True
        if not self.solo_channel or append_one:
            shape_dims = 2
        else:
            shape_dims = 3
        assert len(buf.shape) >= shape_dims
        if len(buf.shape) == shape_dims:
            return self._msgpack.encode(
                [{"items": self.encode_single_image(buf), "append_one": append_one}]
            )
        else:
            image_shape = buf.shape[-shape_dims:]
            items_shape = buf.shape[:-shape_dims]
            items = []
            for i in np.ndindex(items_shape):
                items.append(self.encode_single_image(buf[i]))
            return self._msgpack.encode(
                [
                    {
                        "items": items,
                        "items_shape": items_shape,
                        "image_shape": image_shape,
                        "dtype": str(buf.dtype),
                        "append_one": append_one,
                    }
                ]
            )

    def decode(self, buf, out=None):
        data = self._msgpack.decode(buf)[0]
        if "items_shape" not in data:
            images = self.decode_single_image(data["items"])
        else:
            items = data["items"]
            images = np.zeros(
                data["items_shape"] + data["image_shape"], dtype=data["dtype"]
            )

            for i, index in enumerate(np.ndindex(tuple(data["items_shape"]))):
                images[index] = self.decode_single_image(items[i])

        if data.get("append_one"):
            images = np.reshape(images, images.shape + (1,))
        return images

    def get_config(self):
        return {"id": self.codec_id, "solo_channel": self.solo_channel}

    # def __dict__(self):
    #     return self.get_config()

    @classmethod
    def from_config(cls, config):
        return PngCodec(config["solo_channel"])


numcodecs.register_codec(PngCodec, "png")



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.exceptions import HubException
from hub.store.shape_detector import ShapeDetector
import pytest


def test_shape_detector():
    s = ShapeDetector((10, 10, 10), 10)
    assert str(s.dtype) == "float64"
    assert s.chunks[1:] == (10, 10)


def test_shape_detector_2():
    s = ShapeDetector((10, 10, 10), 10, compressor="png")
    assert str(s.dtype) == "float64"
    assert s.chunks[1:] == (10, 10)


def test_shape_detector_wrong_shape():
    with pytest.raises(HubException):
        ShapeDetector((10, 10, 10), (10, 10, 20))


def test_shape_detector_wrong_shape_2():
    with pytest.raises(AssertionError):
        ShapeDetector((10, 10, 10), 20)


def test_shape_detector_wrong_shape_3():
    with pytest.raises(HubException):
        ShapeDetector((10, 10, None), (10, 10, None))


def test_shape_detector_wrong_chunk_shape():
    with pytest.raises(Exception):
        ShapeDetector((10, 10, 10), (10, 10, 10), (10, 10))


def test_shape_detector_wrong_chunk_value():
    with pytest.raises(Exception):
        ShapeDetector((10, 10, 10), (10, 10, 10), (2, 10, 10))



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import zlib

from .zip import Zip


class Zlib(Zip):
    def __init__(self, compresslevel: int):
        super().__init__(zlib, compresslevel)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import click
from hub import config
from hub.log import configure_logger
from hub.cli.auth import login, logout, register
from hub.version import __version__
from hub.cli.auth import login, logout, register, reporting


@click.group()
@click.option(
    "-h",
    "--host",
    default=f"{config.HUB_REST_ENDPOINT}",
    help="Hub rest endpoint",
)
@click.option("-v", "--verbose", count=True, help="Devel debugging")
def cli(host, verbose):
    configure_logger(verbose)


@click.group()
@click.version_option(__version__)
@click.pass_context
def cli(ctx):
    pass


def add_commands(cli):
    cli.add_command(login)
    cli.add_command(register)
    cli.add_command(logout)
    cli.add_command(cli)
    cli.add_command(reporting)


add_commands(cli)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import gzip

from .zip import Zip


class Gzip(Zip):
    def __init__(self, compresslevel: int):
        super().__init__(gzip, compresslevel)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from collections.abc import MutableMapping
import posixpath

import boto3
import botocore
from botocore.exceptions import ClientError
from s3fs import S3FileSystem

from hub.exceptions import S3Exception
from hub.log import logger


class S3Storage(MutableMapping):
    def __init__(
        self,
        s3fs: S3FileSystem,
        url: str = None,
        public=False,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        aws_session_token=None,
        parallel=25,
        endpoint_url=None,
    ):
        self.s3fs = s3fs
        self.root = {}
        self.url = url
        self.public = public
        self.parallel = parallel
        self.endpoint_url = endpoint_url
        self.bucket = url.split("/")[2]
        self.path = "/".join(url.split("/")[3:])
        if self.bucket == "s3:":
            # FIXME for some reason this is wasabi case here, probably url is something like wasabi://s3://...
            self.bucket = url.split("/")[4]
            self.path = "/".join(url.split("/")[5:])
        self.bucketpath = posixpath.join(self.bucket, self.path)
        self.protocol = "object"

        client_config = botocore.config.Config(
            max_pool_connections=parallel,
        )

        self.client = boto3.client(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            config=client_config,
            endpoint_url=endpoint_url,
        )

        self.resource = boto3.resource(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            config=client_config,
            endpoint_url=endpoint_url,
        )

    def __setitem__(self, path, content):
        try:
            path = posixpath.join(self.path, path)
            content = bytearray(memoryview(content))
            attrs = {
                "Bucket": self.bucket,
                "Body": content,
                "Key": path,
                "ContentType": ("application/octet-stream"),
            }

            self.client.put_object(**attrs)
        except Exception as err:
            logger.error(err)
            raise S3Exception(err)

    def __getitem__(self, path):
        try:
            path = posixpath.join(self.path, path)
            resp = self.client.get_object(
                Bucket=self.bucket,
                Key=path,
            )
            x = resp["Body"].read()
            return x
        except ClientError as err:
            if err.response["Error"]["Code"] == "NoSuchKey":
                raise KeyError(err)
            else:
                raise
        except Exception as err:
            logger.error(err)
            raise S3Exception(err)

    def __delitem__(self, path):
        try:
            path = posixpath.join(self.bucketpath, path)
            self.s3fs.rm(path, recursive=True)
        except Exception as err:
            logger.error(err)
            raise S3Exception(err)

    def __len__(self):
        return len(self.s3fs.ls(self.bucketpath, detail=False, refresh=True))

    def __iter__(self):
        items = self.s3fs.ls(self.bucketpath, detail=False, refresh=True)
        yield from [item[len(self.bucketpath) + 1 :] for item in items]



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import Tuple

import numpy as np

from hub.schema.features import Tensor


class Image(Tensor):
    """| `HubSchema` for images.

    Output: `tf.Tensor` of type `tf.uint8` and shape `[height, width, num_channels]`
    for BMP, JPEG, and PNG images

    Example:
    ----------
    >>> image_tensor = Image(shape=(None, None, 1),
    >>>                      encoding_format='png')
    """

    def __init__(
        self,
        shape: Tuple[int, ...] = (None, None, 3),
        dtype="uint8",
        # TODO Add back encoding_format (probably named compress) when support for png and jpg support will be added
        max_shape: Tuple[int, ...] = None,
        chunks=None,
        compressor="lz4",
    ):
        """| Construct the connector.

        Parameters
        ----------
        shape: tuple of ints or None
            The shape of decoded image:
            (height, width, channels) where height and width can be None.
            Defaults to (None, None, 3).
        dtype: `uint16` or `uint8` (default)
            `uint16` can be used only with png encoding_format
        encoding_format: 'jpeg' or 'png' (default)
             Format to serialize np.ndarray images on disk.
        max_shape : Tuple[int]
            Maximum shape of tensor shape if tensor is dynamic
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks


        Returns
        ----------
        `tf.Tensor` of type `tf.uint8` and shape `[height, width, num_channels]`
        for BMP, JPEG, and PNG images

        Raises
        ----------
        ValueError: If the shape, dtype or encoding formats are invalid

        """
        self._set_dtype(dtype)
        super().__init__(
            shape,
            dtype,
            max_shape=max_shape,
            chunks=chunks,
            compressor=compressor,
        )

    def _set_dtype(self, dtype):
        """Set the dtype."""
        dtype = str(np.dtype(dtype))
        if dtype not in ("uint8", "uint16"):
            raise ValueError(f"Not supported dtype for {self.__class__.__name__}")
        self.dtype = dtype

    def __str__(self):
        out = super().__str__()
        out = "Image" + out[6:]
        return out

    def __repr__(self):
        return self.__str__()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import Tuple

from hub.schema import Tensor


class Video(Tensor):
    """`HubSchema` for videos, encoding frames individually on disk.

    The connector accepts as input a 4 dimensional `uint8` array
    representing a video.

    Returns
    ----------
    Tensor: `uint8` and shape [num_frames, height, width, channels],
         where channels must be 1 or 3
    """

    def __init__(
        self,
        shape: Tuple[int, ...] = None,
        dtype: str = "uint8",
        # TODO Add back encoding_format (probably named compress) when support for png and jpg support will be added
        max_shape: Tuple[int, ...] = None,
        # ffmpeg_extra_args=(),
        chunks=None,
        compressor="lz4",
    ):
        """Initializes the connector.

        Parameters
        ----------

        shape: tuple of ints
            The shape of the video (num_frames, height, width,
            channels), where channels is 1 or 3.
        encoding_format: str
            The video is stored as a sequence of encoded images.
            You can use any encoding format supported by Image.
        dtype: `uint16` or `uint8` (default)

        Raises
        ----------
        ValueError: If the shape, dtype or encoding formats are invalid
        """
        super(Video, self).__init__(
            dtype=dtype,
            shape=shape,
            max_shape=max_shape,
            chunks=chunks,
            compressor=compressor,
        )

    def __str__(self):
        out = super().__str__()
        out = "Video" + out[6:]
        return out

    def __repr__(self):
        return self.__str__()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from fsspec import AbstractFileSystem
import array
from collections.abc import MutableMapping
from azure.storage.blob import BlobServiceClient


class AzureBlobFileSystem(AbstractFileSystem):
    def __init__(
        self,
        account_name: str,
        account_key: str = None,
        connection_string: str = None,
        credential: str = None,
        sas_token: str = None,
        request_session=None,
        socket_timeout: int = None,
        client_id: str = None,
        client_secret: str = None,
        tenant_id: str = None,
    ):

        super().__init__()
        self.account_name = account_name
        self.account_key = account_key
        self.connection_string = connection_string
        self.credential = credential
        self.sas_token = sas_token
        self.request_session = request_session
        self.socket_timeout = socket_timeout
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        if (
            self.credential is None
            and self.account_key is None
            and self.sas_token is None
            and self.client_id is not None
        ):
            self.credential = self._get_credential_from_service_principal()
        self.do_connect()

    def _get_credential_from_service_principal(self):
        """
        Create a Credential for authentication.  This can include a TokenCredential
        client_id, client_secret and tenant_id

        Returns
        -------
        Credential
        """
        from azure.identity import ClientSecretCredential

        sp_token = ClientSecretCredential(
            tenant_id=self.tenant_id,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )
        return sp_token

    def do_connect(self):
        """Connect to the BlobServiceClient, using user-specified connection details.
        Tries credentials first, then connection string and finally account key

        Raises
        ------
        ValueError if none of the connection details are available
        """
        try:
            self.account_url: str = f"https://{self.account_name}.blob.core.windows.net"
            if self.credential is not None:
                self.service_client = BlobServiceClient(
                    account_url=self.account_url, credential=self.credential
                )
            elif self.connection_string is not None:
                self.service_client = BlobServiceClient.from_connection_string(
                    conn_str=self.connection_string
                )
            elif self.account_key is not None:
                self.service_client = BlobServiceClient(
                    account_url=self.account_url, credential=self.account_key
                )
            elif self.sas_token is not None:
                self.service_client = BlobServiceClient(
                    account_url=self.account_url + self.sas_token, credential=None
                )
            else:
                self.service_client = BlobServiceClient(account_url=self.account_url)

        except Exception as e:
            raise ValueError(f"unable to connect to account for {e}")

    def exists(self, path):
        """
        Checks whether the given path exists in the File System

        Returns
        -------
        Boolean
        """
        split_path = path.split("/")
        container_name = split_path[0]
        sub_path = "/".join(split_path[1:])
        container = self.service_client.get_container_client(container_name)
        it = container.list_blobs(name_starts_with=sub_path)
        return len(list(it)) > 0

    def ls(self, path, refresh=True):
        """
        Finds all the files in the given path in the File System
        Returns
        -------
        List of full paths of all files found in given path
        """
        return self.find(path)

    def isfile(self, path):
        """Is this entry file-like?
        Azure fs only stores path to files and not folders. This is always true
        """
        return True

    def find(self, path):
        """
        Finds all the files in the given path in the File System

        Returns
        -------
        List of full paths of all files found in given path
        """
        split_path = path.split("/")
        container_name = split_path[0]
        sub_path = "/".join(split_path[1:])
        container = self.service_client.get_container_client(container_name)
        it = container.list_blobs(name_starts_with=sub_path)
        return [f"{container_name}/{item['name']}" for item in it]

    def rm(self, path, recursive=False, maxdepth=None):
        """Removes all the files in the given path"""
        split_path = path.split("/")
        container_name = split_path[0]
        sub_path = "/".join(split_path[1:])
        container = self.service_client.get_container_client(container_name)
        it = container.list_blobs(name_starts_with=sub_path)
        for item in it:
            container.delete_blob(item)

    def makedirs(self, path, exist_ok=False):
        """Recursively creates directories in path"""
        # in azure empty directories have no meaning, so makedirs not needed
        return

    def get_mapper(self, root, check=False, create=False):
        """Create key-value interface for given root"""
        return FSMap(root, self)

    def upload(self, path, value):
        """Uploads value to the given path"""
        split_path = path.split("/")
        container_name = split_path[0]
        sub_path = "/".join(split_path[1:])
        blob_client = self.service_client.get_blob_client(container_name, sub_path)
        blob_client.upload_blob(value, overwrite=True)

    def download(self, path):
        """Downloads the value from the given path"""
        if not self.exists(path):
            raise KeyError()
        split_path = path.split("/")
        container_name = split_path[0]
        sub_path = "/".join(split_path[1:])
        blob_client = self.service_client.get_blob_client(container_name, sub_path)
        return blob_client.download_blob().readall()

    def cat_file(self, path):
        return self.download(path)

    def pipe_file(self, path, value):
        return self.upload(path, value)


class FSMap(MutableMapping):
    def __init__(self, root, fs, check=False, create=False, missing_exceptions=None):
        self.fs = fs
        self.root = root
        if missing_exceptions is None:
            missing_exceptions = (
                FileNotFoundError,
                IsADirectoryError,
                NotADirectoryError,
            )
        self.missing_exceptions = missing_exceptions
        if check and not create:
            if not self.fs.exists(root):
                raise ValueError(
                    "Path %s does not exist. Create "
                    " with the ``create=True`` keyword" % root
                )

    def clear(self):
        """Remove all keys below root - empties out mapping"""
        self.fs.rm(self.root, True)

    def _key_to_str(self, key):
        """Generate full path for the key"""
        if isinstance(key, (tuple, list)):
            key = str(tuple(key))
        else:
            key = str(key)
        return "/".join([self.root, key]) if self.root else key

    def _str_to_key(self, s):
        """Strip path of to leave key name"""
        return s[len(self.root) :].lstrip("/")

    def __getitem__(self, key, default=None):
        """Retrieve data"""
        k = self._key_to_str(key)
        try:
            result = self.fs.download(k)
        except self.missing_exceptions:
            if default is not None:
                return default
            raise KeyError(key)
        return result

    def __setitem__(self, key, value):
        """Store value in key"""
        key = self._key_to_str(key)
        if isinstance(value, array.array) or isinstance(value, memoryview):
            value = bytearray(value)
        self.fs.upload(key, value)

    def __iter__(self):
        """iterating over the structure"""
        return (self._str_to_key(x) for x in self.fs.find(self.root))

    def __len__(self):
        """returns length of the structure"""
        return len(self.fs.find(self.root))

    def __delitem__(self, key):
        """Remove key"""
        self.fs.rm(self._key_to_str(key))

    def __contains__(self, key):
        """Does key exist in mapping?"""
        path = self._key_to_str(key)
        return self.fs.exists(path)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import os
import hub
from outdated import check_outdated
import subprocess
import pkg_resources

from hub.exceptions import HubException
from hub.log import logger


def get_cli_version():
    return "1.0.0"



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.store.s3_storage import S3Storage
import copyreg
from hub.store.s3_storage import S3Storage


def save_s3_storage(obj):
    return obj.__class__, (
        obj.s3fs,
        obj.url,
        obj.public,
        None,
        None,
        None,
        obj.parallel,
        obj.endpoint_url,
    )


copyreg.pickle(S3Storage, save_s3_storage)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import io

import numpy as np

from .base import Base


class Zip(Base):
    def __init__(self, compressor, compresslevel: int):
        self._compressor = compressor
        self._compresslevel = compresslevel

    def encode(self, array: np.ndarray) -> bytes:
        with io.BytesIO() as f:
            np.save(f, array, allow_pickle=True)
            return self._compressor.compress(f.getvalue(), self._compresslevel)

    def decode(self, content: bytes) -> np.ndarray:
        with io.BytesIO(self._compressor.decompress(content)) as f:
            return np.load(f, allow_pickle=True)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.exceptions import DynamicTensorShapeException
import posixpath

import numpy as np
import fsspec
from zarr.creation import create

from hub.store.dynamic_tensor import DynamicTensor
from hub.store.store import StorageMapWrapperWithCommit


def create_store(path: str, overwrite=True):
    fs: fsspec.AbstractFileSystem = fsspec.filesystem("file")
    if fs.exists(path) and overwrite:
        fs.rm(path, recursive=True)
    dynpath = posixpath.join(path, "--dynamic--")
    if not fs.exists(dynpath):
        fs.makedirs(dynpath)
    mapper = fs.get_mapper(path)
    mapper["--dynamic--/hello.txt"] = bytes("Hello World", "utf-8")
    return StorageMapWrapperWithCommit(mapper)


def test_read_and_append_modes():
    t = DynamicTensor(
        create_store("./data/test/test_read_and_append_modes"),
        mode="a",
        shape=(5, 100, 100),
        max_shape=(5, 100, 100),
        dtype="int32",
    )
    t[0, 80:, 80:] = np.ones((20, 20), dtype="int32")
    assert t[0, -5, 90:].tolist() == [1] * 10
    t.flush()
    t.close()

    t = DynamicTensor(
        create_store("./data/test/test_read_and_append_modes", overwrite=False),
        mode="r",
    )
    t.get_shape(0) == (100, 100)
    assert t[0, -5, 90:].tolist() == [1] * 10
    t.close()


def test_dynamic_tensor():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor"),
        mode="w",
        shape=(5, 100, 100),
        max_shape=(5, 100, 100),
        dtype="int32",
    )
    t[0, 80:, 80:] = np.ones((20, 20), dtype="int32")
    assert t[0, -5, 90:].tolist() == [1] * 10


def test_dynamic_tensor_shape_none():
    try:
        DynamicTensor(
            create_store("./data/test/test_dynamic_tensor_shape_none"),
            mode="w",
            dtype="int32",
        )
    except TypeError as ex:
        assert "shape cannot be none" in str(ex)


def test_dynamic_tensor_2():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor_2"),
        mode="w",
        shape=(5, None, None),
        max_shape=(5, 100, 100),
        dtype="int32",
    )
    t[0] = np.ones((10, 10), dtype="int32")
    assert t[0, 5].tolist() == [1] * 10
    assert t[0, 5, :].tolist() == [1] * 10
    t[0, 6] = 2 * np.ones((20,), dtype="int32")
    assert t[0, 5, :].tolist() == [1] * 10 + [0] * 10
    assert t.get_shape(0).tolist() == [10, 20]
    assert t.get_shape(slice(0, 1)).tolist() == [1, 10, 20]


def test_dynamic_tensor_3():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor_3"),
        mode="w",
        shape=(5, None, None, None),
        max_shape=(5, 100, 100, 100),
        dtype="int32",
    )
    t[0, 5] = np.ones((20, 30), dtype="int32")
    t[0, 6:8, 5:9] = 5 * np.ones((2, 4, 30), dtype="int32")
    assert t[0, 5, 7].tolist() == [1] * 30
    assert t[0, 7, 8].tolist() == [5] * 30


def test_dynamic_tensor_shapes():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor_5"),
        mode="w",
        shape=(5, None, None),
        max_shape=(5, 100, 100),
        dtype="int32",
    )
    t[0] = np.ones((5, 10), dtype="int32")
    t[0, 6] = 2 * np.ones((20,), dtype="int32")
    assert t[0, -1].tolist() == [2] * 20
    t.close()


def test_dynamic_tensor_4():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor_6"),
        mode="w",
        shape=(5, None, None, None),
        max_shape=(5, 100, 100, 10),
        dtype="int32",
    )
    t[0, 6:8] = np.ones((2, 20, 10), dtype="int32")
    assert (t[0, 6:8] == np.ones((2, 20, 10), dtype="int32")).all()


if __name__ == "__main__":
    test_read_and_append_modes()
    # test_chunk_iterator()
    # test_dynamic_tensor_shapes()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import sys
from hub import Dataset
from hub.api.datasetview import DatasetView
from hub.utils import batchify
from hub.compute import Transform
from typing import Iterable, Iterator
from hub.exceptions import ModuleNotInstalledException
from hub.api.sharded_datasetview import ShardedDatasetView
import hub
from hub.api.dataset_utils import get_value, str_to_int


def empty_remote(template, **kwargs):
    """
    remote template
    """

    def wrapper(func):
        def inner(**kwargs):
            return func

        return inner

    return wrapper


try:
    import ray

    remote = ray.remote
except Exception:
    remote = empty_remote


class RayTransform(Transform):
    def __init__(self, func, schema, ds, scheduler="ray", workers=1, **kwargs):
        super(RayTransform, self).__init__(
            func, schema, ds, scheduler="single", workers=workers, **kwargs
        )
        self.workers = workers
        if "ray" not in sys.modules:
            raise ModuleNotInstalledException("ray")

        if not ray.is_initialized():
            ray.init(local_mode=True)

    @remote
    def _func_argd(_func, index, _ds, schema, kwargs):
        """
        Remote wrapper for user defined function
        """

        if isinstance(_ds, (Dataset, DatasetView)) and isinstance(_ds.indexes, int):
            _ds.indexes = [_ds.indexes]

        item = _ds[index]
        if isinstance(item, DatasetView) or isinstance(item, Dataset):
            item = item.compute()

        item = _func(0, item)
        item = Transform._flatten_dict(item, schema=schema)

        return list(item.values())

    def store(
        self,
        url: str,
        token: dict = None,
        length: int = None,
        ds: Iterable = None,
        progressbar: bool = True,
        public: bool = True,
    ):
        """
        The function to apply the transformation for each element in batchified manner

        Parameters
        ----------
        url: str
            path where the data is going to be stored
        token: str or dict, optional
            If url is refering to a place where authorization is required,
            token is the parameter to pass the credentials, it can be filepath or dict
        length: int
            in case shape is None, user can provide length
        ds: Iterable
        progressbar: bool
            Show progress bar
        public: bool, optional
            only applicable if using hub storage, ignored otherwise
            setting this to False allows only the user who created it to access the dataset and
            the dataset won't be visible in the visualizer to the public

        Returns
        ----------
        ds: hub.Dataset
            uploaded dataset
        """
        _ds = ds or self.base_ds

        num_returns = len(self._flatten_dict(self.schema, schema=self.schema).keys())
        results = [
            self._func_argd.options(num_returns=num_returns).remote(
                self.call_func, el, _ds, schema=self.schema, kwargs=self.kwargs
            )
            for el in range(len(_ds))
        ]

        if num_returns == 1:
            results = [[r] for r in results]

        results = self._split_list_to_dicts(results)

        ds = self.upload(
            results, url=url, token=token, progressbar=progressbar, public=public
        )
        return ds

    @remote
    def upload_chunk(i_batch, key, ds):
        """
        Remote function to upload a chunk
        Returns the shape of dynamic tensor to upload all in once after upload is completed

        Parameters
        ----------
        i_batch: Tuple
            Tuple composed of (index, batch)
        key: str
            Key of the tensor
        ds:
            Dataset to set to upload
        Returns
        ----------
        (key, slice_, shape) to set the shape later

        """
        i, batch = i_batch
        if not isinstance(batch, dict) and isinstance(batch[0], ray.ObjectRef):
            batch = ray.get(batch)
            # FIXME an ugly hack to unwrap elements with a schema that has one tensor
            num_returns = len(
                Transform._flatten_dict(ds.schema.dict_, schema=ds.schema.dict_).keys()
            )
            if num_returns == 1:
                batch = [item for sublist in batch for item in sublist]

        shape = None
        length = len(batch)

        slice_ = slice(i * length, (i + 1) * length)
        if ds[key].is_dynamic:
            # Sometimes ds._tensor slice_ gets out of the shape value
            shape = ds._tensors[f"/{key}"].get_shape_from_value([slice_], batch)
        ds[key, slice_] = batch

        return (key, [slice_], shape)

    def upload(
        self,
        results,
        url: str,
        token: dict,
        progressbar: bool = True,
        public: bool = True,
    ):
        """Batchified upload of results.
        For each tensor batchify based on its chunk and upload.
        If tensor is dynamic then still upload element by element.

        Parameters
        ----------
        dataset: hub.Dataset
            Dataset object that should be written to
        results:
            Output of transform function
        progressbar: bool
        public: bool, optional
            only applicable if using hub storage, ignored otherwise
            setting this to False allows only the user who created it to access the dataset and
            the dataset won't be visible in the visualizer to the public
        Returns
        ----------
        ds: hub.Dataset
            Uploaded dataset
        """
        if len(list(results.values())) == 0:
            shape = (0,)
        else:
            shape = (len(list(results.values())[0]),)

        ds = Dataset(
            url,
            mode="w",
            shape=shape,
            schema=self.schema,
            token=token,
            cache=False,
            public=public,
        )

        tasks = []
        for key, value in results.items():

            length = ds[key].chunksize[0]
            value = get_value(value)
            value = str_to_int(value, ds.tokenizer)
            batched_values = batchify(value, length)
            chunk_id = list(range(len(batched_values)))
            index_batched_values = list(zip(chunk_id, batched_values))

            ds._tensors[f"/{key}"].disable_dynamicness()

            results = [
                self.upload_chunk.remote(el, key=key, ds=ds)
                for el in index_batched_values
            ]
            tasks.extend(results)

        results = ray.get(tasks)
        self.set_dynamic_shapes(results, ds)
        ds.commit()
        return ds

    def set_dynamic_shapes(self, results, ds):
        """
        Sets shapes for dynamic tensors after the dataset is uploaded

        Parameters
        ----------
        results: Tuple
            results from uploading each chunk which includes (key, slice, shape) tuple
        ds:
            Dataset to set the shapes to
        Returns
        ----------
        """

        shapes = {}
        for (key, slice_, value) in results:
            if not ds[key].is_dynamic:
                continue

            if key not in shapes:
                shapes[key] = []
            shapes[key].append((slice_, value))

        for key, value in shapes.items():
            ds._tensors[f"/{key}"].enable_dynamicness()

            for (slice_, shape) in shapes[key]:
                ds._tensors[f"/{key}"].set_dynamic_shape(slice_, shape)


class TransformShard:
    def __init__(self, ds, func, schema, kwargs):

        if isinstance(ds, Dataset) or isinstance(ds, DatasetView):
            ds.squeeze_dim = False

        self._ds = ds
        self._func = func
        self.schema = schema
        self.kwargs = kwargs
        self.token = None

    def __call__(self, ids):
        """
        For each shard, transform each sample and then store inside shared memory of ray
        """
        for index in ids:
            item = self._ds[index]
            if isinstance(item, DatasetView) or isinstance(item, Dataset):
                item = item.compute()

            items = self._func(0, item)
            if not isinstance(items, list):
                items = [items]

            for item in items:
                yield Transform._flatten_dict(item, schema=self.schema)


class RayGeneratorTransform(RayTransform):
    def store(
        self,
        url: str,
        token: dict = None,
        length: int = None,
        ds: Iterable = None,
        progressbar: bool = True,
        public: bool = True,
    ):
        """
        The function to apply the transformation for each element by sharding the dataset

        Parameters
        ----------
        url: str
            path where the data is going to be stored
        token: str or dict, optional
            If url is refering to a place where authorization is required,
            token is the parameter to pass the credentials, it can be filepath or dict
        length: int
            in case shape is None, user can provide length
        ds: Iterable
        progressbar: bool
            Show progress bar
        public: bool, optional
            only applicable if using hub storage, ignored otherwise
            setting this to False allows only the user who created it to access the dataset and
            the dataset won't be visible in the visualizer to the public
        Returns
        ----------
        ds: hub.Dataset
            uploaded dataset
        """
        _ds = ds or self.base_ds

        results = ray.util.iter.from_range(len(_ds), num_shards=self.workers).transform(
            TransformShard(
                ds=_ds, func=self.call_func, schema=self.schema, kwargs=self.kwargs
            )
        )

        @remote
        def upload_shard(i, shard_results):
            """
            Create a seperate dataset per shard
            """
            shard_results = self._split_list_to_dicts(shard_results)

            if len(shard_results) == 0 or len(list(shard_results.values())[0]) == 0:
                return None

            ds = self.upload(
                shard_results,
                url=f"{url}_shard_{i}",
                token=token,
                progressbar=progressbar,
                public=public,
            )
            return ds

        work = [
            upload_shard.remote(i, shard) for i, shard in enumerate(results.shards())
        ]

        datasets = ray.get(work)
        datasets = [d for d in datasets if d]
        ds = self.merge_sharded_dataset(datasets, url, token=token)
        return ds

    def merge_sharded_dataset(
        self,
        datasets,
        url,
        token=None,
    ):
        """
        Creates a sharded dataset and then applies ray transform to upload it
        """
        sharded_ds = ShardedDatasetView(datasets)

        @hub.transform(schema=self.schema, scheduler="ray")
        def transform_identity(sample):
            return sample

        ds = transform_identity(sharded_ds).store(
            url,
            token=token,
        )

        return ds



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from concurrent.futures.thread import ThreadPoolExecutor

import boto3
import pytest
from s3fs import S3FileSystem
import cloudpickle
import numpy as np

from hub.store.s3_storage import S3Storage
from hub.utils import s3_creds_exist


def create_client():
    return boto3.client("s3")


BYTE_DATA = bytes("world2", "utf-8")
NUMPY_ARR = np.array([[4, 512, 512]])


@pytest.mark.skipif(not s3_creds_exist(), reason="Requires s3 credentials")
def test_s3_storage():
    _storage = S3FileSystem()
    storage = S3Storage(_storage, "s3://snark-test/test_s3_storage1")

    storage["hello"] = BYTE_DATA
    storage["numpy"] = NUMPY_ARR
    assert storage["hello"] == BYTE_DATA
    assert storage["numpy"] == bytearray(memoryview(NUMPY_ARR))
    assert len(storage) == 2
    assert list(storage) == ["hello", "numpy"]
    del storage["hello"]
    assert len(storage) == 1


@pytest.mark.skipif(not s3_creds_exist(), reason="Requires s3 credentials")
def test_s3_storage_pickability():
    _storage = S3FileSystem()
    storage = S3Storage(_storage, "s3://snark-test/test_s3_storage")

    cloudpickle.dumps(storage)


if __name__ == "__main__":
    test_s3_storage()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.schema.features import Tensor, SchemaDict
from hub.schema.image import Image
from hub.schema.class_label import ClassLabel
from hub.schema.polygon import Polygon
from hub.schema.audio import Audio
from hub.schema.bbox import BBox
from hub.schema.mask import Mask
from hub.schema.segmentation import Segmentation
from hub.schema.sequence import Sequence
from hub.schema.video import Video
from hub.schema.text import Text


def _get_compressor(inp):
    return inp.get("compressor") or "default"


def deserialize(inp):
    if isinstance(inp, dict):
        if inp["type"] == "Audio":
            return Audio(
                shape=tuple(inp["shape"]),
                dtype=deserialize(inp["dtype"]),
                file_format=inp["file_format"],
                sample_rate=inp["sample_rate"],
                max_shape=tuple(inp["max_shape"]),
                chunks=inp["chunks"],
                compressor=_get_compressor(inp),
            )
        elif inp["type"] == "BBox":
            return BBox(
                dtype=deserialize(inp["dtype"]),
                chunks=inp["chunks"],
                compressor=_get_compressor(inp),
            )
        elif inp["type"] == "ClassLabel":
            if inp["_names"] is not None:
                return ClassLabel(
                    names=inp["_names"],
                    chunks=inp["chunks"],
                    compressor=_get_compressor(inp),
                )
            else:
                return ClassLabel(
                    num_classes=inp["_num_classes"],
                    chunks=inp["chunks"],
                    compressor=_get_compressor(inp),
                )
        elif inp["type"] == "SchemaDict" or inp["type"] == "FeatureDict":
            d = {}
            for k, v in inp["items"].items():
                d[k] = deserialize(v)
            return SchemaDict(d)
        elif inp["type"] == "Image":
            return Image(
                shape=tuple(inp["shape"]),
                dtype=deserialize(inp["dtype"]),
                # TODO uncomment back when image encoding will be added
                # encoding_format=inp["encoding_format"],
                max_shape=tuple(inp["max_shape"]),
                chunks=inp["chunks"],
                compressor=_get_compressor(inp),
            )
        elif inp["type"] == "Mask":
            return Mask(
                shape=tuple(inp["shape"]),
                max_shape=tuple(inp["max_shape"]),
                chunks=inp["chunks"],
                compressor=_get_compressor(inp),
            )
        elif inp["type"] == "Polygon":
            return Polygon(
                shape=tuple(inp["shape"]),
                max_shape=tuple(inp["max_shape"]),
                dtype=deserialize(inp["dtype"]),
                chunks=inp["chunks"],
                compressor=_get_compressor(inp),
            )
        elif inp["type"] == "Segmentation":
            class_labels = deserialize(inp["class_labels"])
            if class_labels._names is not None:
                return Segmentation(
                    shape=tuple(inp["shape"]),
                    dtype=deserialize(inp["dtype"]),
                    names=class_labels._names,
                    max_shape=tuple(inp["max_shape"]),
                    chunks=inp["chunks"],
                    compressor=_get_compressor(inp),
                )
            else:
                return Segmentation(
                    shape=tuple(inp["shape"]),
                    dtype=deserialize(inp["dtype"]),
                    num_classes=class_labels._num_classes,
                    max_shape=tuple(inp["max_shape"]),
                    chunks=inp["chunks"],
                    compressor=_get_compressor(inp),
                )
        elif inp["type"] == "Sequence":
            return Sequence(
                shape=tuple(inp["shape"]),
                dtype=deserialize(inp["dtype"]),
                max_shape=tuple(inp["max_shape"]),
                chunks=inp["chunks"],
                compressor=_get_compressor(inp),
            )
        elif inp["type"] == "Tensor":
            return Tensor(
                tuple(inp["shape"]),
                deserialize(inp["dtype"]),
                max_shape=tuple(inp["max_shape"]),
                chunks=inp["chunks"],
                compressor=_get_compressor(inp),
            )
        elif inp["type"] == "Text":
            return Text(
                tuple(inp["shape"]),
                deserialize(inp["dtype"]),
                max_shape=tuple(inp["max_shape"]),
                chunks=inp["chunks"],
                compressor=_get_compressor(inp),
            )
        elif inp["type"] == "Video":
            return Video(
                shape=tuple(inp["shape"]),
                dtype=deserialize(inp["dtype"]),
                # TODO uncomment back when image encoding will be added
                # encoding_format=inp["encoding_format"],
                max_shape=tuple(inp["max_shape"]),
                chunks=inp["chunks"],
                compressor=_get_compressor(inp),
            )
    else:
        return inp



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import os

from setuptools import find_packages, setup

project = "hub"
VERSION = "1.2.3"

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md")) as f:
    long_description = f.read()

with open(os.path.join(this_directory, "requirements.txt")) as f:
    requirements = f.readlines()

setup(
    name=project,
    version=VERSION,
    description="Activeloop Hub",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Snark AI Inc.",
    author_email="support@activeloop.ai",
    license="MPL 2.0",
    url="https://github.com/activeloopai/Hub",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    keywords="snark-hub",
    project_urls={
        "Documentation": "https://docs.activeloop.ai/",
        "Source": "https://github.com/activeloopai/Hub",
    },
    classifiers=[
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires=">=3.6",
    install_requires=requirements,
    setup_requires=[],
    dependency_links=[],
    entry_points={
        "console_scripts": [
            "activeloop = hub.cli.command:cli",
            "activeloop-local = hub.cli.local:cli",
            "activeloop-dev = hub.cli.dev:cli",
            "hub = hub.cli.command:cli",
        ]
    },
    tests_require=["pytest", "mock>=1.0.1"],
)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import io

import numpy as np

from .base import Base


class Default(Base):
    def __init__(self):
        super().__init__()

    def encode(self, array: np.ndarray) -> bytes:
        with io.BytesIO() as f:
            np.save(f, array, allow_pickle=True)
            return f.getvalue()

    def decode(self, bytes_: bytes) -> np.ndarray:
        with io.BytesIO(bytes_) as f:
            return np.load(f, allow_pickle=True)
